import collections
import dataclasses
import datetime
import traceback
import types
import typing

@dataclasses.dataclass
class ExceptionInfo:
    ''' A dataclass containing an exception, traceback, and approximate timestamp '''
    exception: Exception
    traceback: types.TracebackType
    time_stamp: datetime.datetime

    def __str__(self):
        ret_str = f'{type(self.exception).__name__}: thrown at {self.time_stamp}\n'
        ret_str += '  ' + str(self.exception).replace('\n', '\n  ').rstrip(' ')
        ret_str += '    ' + ('\n'.join(traceback.format_tb(self.traceback))).replace('\n', '\n   ').rstrip(' ')
        return ret_str

class CombinedException(Exception):
    ''' An exception that will be raised containing the currently eatten exceptions of the same type '''
    def __init__(self, exception_infos: typing.List[ExceptionInfo]):
        self.exception_infos = exception_infos

    def __str__(self):
        return 'CombinedException:\n  ' + ('\n'.join([str(a) for a in self.exception_infos])).replace('\n', '\n  ').rstrip(' ')


class PyNom:
    # A special variable that corresponds with an exception type leading to all exceptions being eatten even if not listed in exception_types_to_eat
    ALL_EXCEPTIONS = 'ALL_EXCEPTIONS'

    def __init__(self, exception_types_to_eat: list,
                       max_to_eat_before_throw_up: int,
                       throw_up_action:typing.Union[typing.Callable, None]=None):
        ''' Instantiate a PyNom object
            exception_types_to_eat: A list of exception types that PyNom should eat.
                Note that if PyNom.ALL_EXCEPTIONS is given, all exceptions will be eaten.
                    Note 2: Passing the max_to_eat_before_throw_up for a given exception type will still lead to a throw up.
            max_to_eat_before_throwing_up: The max exceptions of a given type to eat before throwing up
            throw_up_action: A callable to call when throwing up. The callable should take one arg (a CombinedException).
                If None is given, raise the CombinedException.
        '''
        self.exception_types_to_eat = exception_types_to_eat if isinstance(exception_types_to_eat, (list, set)) else [exception_types_to_eat]
        self.max_to_eat_before_throw_up = max_to_eat_before_throw_up
        self.throw_up_action = throw_up_action
        self._exception_information = collections.defaultdict(list)

    def _is_eating_all_exceptions(self):
        return self.ALL_EXCEPTIONS in self.exception_types_to_eat

    def __enter__(self):
        return self

    def __exit__(self, typ, value, traceback):
        if typ not in self.exception_types_to_eat and not self._is_eating_all_exceptions():
            # about to raise a not eaten exception
            return

        now = datetime.datetime.now()
        self._exception_information[typ].append(ExceptionInfo(value,
                                                              traceback,
                                                              now))

        if len(self._exception_information[typ]) > self.max_to_eat_before_throw_up:
            # define class here so it subclasses the type of the raised exception
            class FullCombinedException(CombinedException, typ):
                pass

            raise_list = self._exception_information[typ]
            self._exception_information[typ] = []

            ex = FullCombinedException(raise_list)
            if self.throw_up_action is None:
                raise ex
            else:
                self.throw_up_action(ex)

        # if we get here, do not throw anything
        return True


import pytest
from . import PyNom, CombinedException, ExceptionInfo

def test_pynom_allows_a_single_exception_type():
    pynom = PyNom(ZeroDivisionError, 3)
    assert pynom.exception_types_to_eat[0] == ZeroDivisionError
    assert len(pynom.exception_types_to_eat) == 1

def test_simple_pynom():
    pynom = PyNom([ZeroDivisionError], 3)

    for i in range(3):
        with pynom:
            1 / 0

    assert len(pynom._exception_information[ZeroDivisionError]) == 3

    try:
        with pynom:
            1/ 0
    except Exception as ex:
        assert isinstance(ex, ZeroDivisionError)
        assert isinstance(ex, CombinedException)
    else:
        assert 0, 'pynom did not raise'

    assert len(pynom._exception_information[ZeroDivisionError]) == 0

def test_pynom_custom_throw_up_action():
    def throw_up_action(ex):
        assert isinstance(ex, CombinedException)
        assert all([isinstance(a, ExceptionInfo) for a in ex.exception_infos])
        throw_up_action.val = True
        raise ex

    pynom = PyNom([ValueError], 1, throw_up_action)

    with pynom:
        raise ValueError("herro")

    with pytest.raises(ValueError):
        with pynom:
            raise ValueError('i am the breaking point')

    assert throw_up_action.val, 'throw_up_action was not called?'

def test_pynom_can_eat_all_exceptions():
    pynom = PyNom([PyNom.ALL_EXCEPTIONS], 1)

    with pynom:
        raise ValueError

    with pynom:
        raise EnvironmentError

    with pynom:
        raise Exception

    with pynom:
        raise TypeError

    with pytest.raises(ValueError):
        with pynom:
            raise ValueError

def test_pynom_wont_eat_not_listed_exception():
    pynom = PyNom([ValueError], 1)

    with pytest.raises(EnvironmentError):
        with pynom:
            raise EnvironmentError



from setuptools import setup
import os
import sys

THIS_FOLDER = os.path.abspath(os.path.dirname(__file__))

def getVersion():
    with open(os.path.join(THIS_FOLDER, 'pynom', '__init__.py'), 'r') as f:
        text = f.read()

    for line in text.splitlines():
        if line.startswith('__version__'):
            version = line.split('=', 1)[1].replace('\'', '').replace('"', '')
            return version.strip()

    raise EnvironmentError("Unable to find __version__!")

setup(
    name='pynom',
    author='csm10495',
    author_email='csm10495@gmail.com',
    url='http://github.com/csm10495/pynom',
    version=getVersion(),
    packages=['pynom'],
    license='MIT License',
    python_requires='>=3,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*,!=3.6.*',
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.md')).read(),
    long_description_content_type="text/markdown",
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    include_package_data = True,
    install_requires=['pytest'],
)